{*******************************************************}
{                                                       }
{       TSRVDxSkinAdapter                               }
{       Copyright (c) 2011 Marek Mittmann               }
{       marek@progtech.pl                               }
{                                                       }
{*******************************************************}

unit SRVDxSkin;

interface

uses
  Classes, Types, Graphics, SclRView, cxGraphics, cxLookAndFeelPainters, cxLookAndFeels,
  SRVSkinManager, SRVSBar, SclRVRuler, RVRuler;

type
  TSRVDxSkinAdapter = class(TComponent)
  private
    FCxCanvasCache: TcxCanvas;
    FSavLookAndFeelChanged: TcxLookAndFeelChangedEvent;
    FDrawEditorFrame: Boolean;
    function GetCxCanvas(Canvas: TCanvas): TcxCanvas;
    function GetPainter: TcxCustomLookAndFeelPainter;
    procedure ReleaseCxCanvas(Canvas: TcxCanvas);
    procedure SRichViewPaint(Sender: TSRichViewEdit; Canvas: TCanvas; Prepaint: Boolean;
      PaintRect: TRect);
    function SRVDrawStateToCxButtonState(State: TSRVDrawStates): TcxButtonState;
    procedure DrawScrollBarArea(Sender: TSRVScrollBar; Canvas: TCanvas; ARect, PaintRect: TRect;
      isFirstArea: Boolean; State: TSRVDrawStates; var DoDefault: Boolean);
    procedure DrawScrollBarDecButton(Sender: TSRVScrollBar; Canvas: TCanvas; ARect,
      PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
    procedure DrawScrollBarIncButton(Sender: TSRVScrollBar; Canvas: TCanvas; ARect,
      PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
    procedure DrawScrollBarThumbTab(Sender: TSRVScrollBar; Canvas: TCanvas; ARect,
      PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
    procedure LookAndFeelChanged(Sender: TcxLookAndFeel; AChangedValues: TcxLookAndFeelValues);
    procedure SetLookAndFeelChangedHandler(Attach: Boolean);
    procedure SetSRichViewPaintHandlers(SRichView: TSRichViewEdit; Attach: Boolean);
    procedure SetPaintHandlers(Attach: Boolean);
    procedure AdjustSRichView(SRichView: TSRichViewEdit);
    procedure AdjustRuler(Ruler: TSclRVRuler);
    procedure AdjustRulerItemSelector(Selector: TRVRulerItemSelector);
    procedure AdjustControls;
    procedure SetDrawEditorFrame(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Refresh;
  published
    property DrawEditorFrame: Boolean read FDrawEditorFrame write SetDrawEditorFrame default True;
  end;


implementation

uses
  Windows, Forms, Math, Ruler, Controls;

type
  TSRVSBarAccess = class(TSRVScrollBar)
  end;


{ TSRVDxSkinAdapterEditor }

procedure TSRVDxSkinAdapter.AdjustControls;
var
  I: Integer;
  Component: TComponent;
begin
  if (Owner <> nil) and not (csDestroying in ComponentState) then
  begin
    for I := 0 to Owner.ComponentCount - 1 do
    begin
      Component := Owner.Components[I];
      if Component is TSRichViewEdit then
        AdjustSRichView(TSRichViewEdit(Component))
      else if Component is TSclRVRuler then
        AdjustRuler(TSclRVRuler(Component))
      else if Component is TRVRulerItemSelector then
        AdjustRulerItemSelector(TRVRulerItemSelector(Component));
    end;
  end;
end;

procedure TSRVDxSkinAdapter.AdjustRuler(Ruler: TSclRVRuler);
var
  Painter: TcxCustomLookAndFeelPainter;
begin
  Painter := GetPainter();
  if Painter <> nil then
  begin
    Ruler.SkinType := stSkin2;
    Ruler.Color := Painter.LayoutControlEmptyAreaColor;
    Ruler.Font.Color := Painter.DefaultHeaderBackgroundColor;
    Ruler.RulerColor := Painter.DefaultHeaderColor;
    Ruler.RulerColorPageEnd := Painter.LayoutControlEmptyAreaColor;
    Ruler.MarginColor := Painter.DefaultHeaderBackgroundColor;
    Ruler.MarginSaturation := 800;
    Ruler.MarginSettings.GripColor := Painter.LayoutControlEmptyAreaColor;
    Ruler.IndentColor := Painter.DefaultHeaderBackgroundColor;
    Ruler.IndentSaturation := 800;
    Ruler.ParentBackground := False;
    Ruler.TableEditor.BackGroundColor := Ruler.RulerColor;
    Ruler.TableEditor.ForeGroundColor := Ruler.Font.Color;
  end;
end;

procedure TSRVDxSkinAdapter.AdjustRulerItemSelector(Selector: TRVRulerItemSelector);
var
  Painter: TcxCustomLookAndFeelPainter;
begin
  Painter := GetPainter();
  if Painter <> nil then
  begin
    Selector.BevelInner := bvNone;
    Selector.BevelHighLightColor := Painter.DefaultHeaderBackgroundColor;
    Selector.BevelShadowColor := Painter.DefaultHeaderBackgroundColor;
    Selector.Color := Painter.DefaultHeaderColor;
    Selector.ParentBackground := False;
  end;
end;

procedure TSRVDxSkinAdapter.AdjustSRichView(SRichView: TSRichViewEdit);
var
  Painter: TcxCustomLookAndFeelPainter;
  Borders: TRect;
begin
  Painter := GetPainter();
  if Painter <> nil then
  begin
    Borders := Painter.PrintPreviewPageBordersWidth;
    SRichView.Color := Painter.LayoutControlEmptyAreaColor;
    SRichView.PageProperty.MinPageBreakHeight := Max(Min(Borders.Top + Borders.Bottom, 20), 5);
    SRichView.PageProperty.MaxPageBreakHeight := Max(Borders.Top + Borders.Bottom, 25);
    SRichView.PageProperty.PageBreakHeight := 20;
  end;
end;

constructor TSRVDxSkinAdapter.Create(AOwner: TComponent);
begin
  inherited;
  FDrawEditorFrame := True;
  SetLookAndFeelChangedHandler(True);
  SetPaintHandlers(True);
  AdjustControls;
end;

destructor TSRVDxSkinAdapter.Destroy;
begin
  SetLookAndFeelChangedHandler(False);
  SetPaintHandlers(False);
  FCxCanvasCache.Free;
  FCxCanvasCache := nil;
  inherited;
end;

procedure TSRVDxSkinAdapter.DrawScrollBarArea(Sender: TSRVScrollBar; Canvas: TCanvas;
  ARect, PaintRect: TRect; isFirstArea: Boolean; State: TSRVDrawStates; var DoDefault: Boolean);
var
  Painter: TcxCustomLookAndFeelPainter;
  CxCanvas: TcxCanvas;
  ScrlPart: TcxScrollBarPart;
begin
  Painter := GetPainter();
  if (Painter <> nil) then
  begin
    CxCanvas := GetCxCanvas(Canvas);
    if (CxCanvas <> nil) then
    begin
      if isFirstArea then
        ScrlPart := sbpPageUp
      else
        ScrlPart := sbpPageDown;
      Painter.DrawScrollBarPart(CxCanvas, Sender.Kind = sbHorizontal, ARect, ScrlPart,
        SRVDrawStateToCxButtonState(State));
      ReleaseCxCanvas(CxCanvas);
      DoDefault := False;
    end;
  end;
end;

procedure TSRVDxSkinAdapter.DrawScrollBarDecButton(Sender: TSRVScrollBar;
  Canvas: TCanvas; ARect, PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
var
  Painter: TcxCustomLookAndFeelPainter;
  R: TRect;
  Bmp: TcxBitmap32;
begin
  Bmp := nil;
  try
    Painter := GetPainter();
    if (Painter <> nil) then
    begin
      R := ARect;
      OffsetRect(R, -R.Left, -R.Top);
      Bmp := TcxBitmap32.CreateSize(R);
      Painter.DrawScrollBarPart(Bmp.CxCanvas, Sender.Kind = sbHorizontal, R, sbpLineUp,
        SRVDrawStateToCxButtonState(State));
      Canvas.Draw(ARect.Left, ARect.Top, Bmp);
      DoDefault := False;
    end;
  except
  end;
  Bmp.Free;
end;

procedure TSRVDxSkinAdapter.DrawScrollBarIncButton(Sender: TSRVScrollBar;
  Canvas: TCanvas; ARect, PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
var
  Painter: TcxCustomLookAndFeelPainter;
  R: TRect;
  Bmp: TcxBitmap32;
begin
  Bmp := nil;
  try
    Painter := GetPainter();
    if (Painter <> nil) then
    begin
      R := ARect;
      OffsetRect(R, -R.Left, -R.Top);
      Bmp := TcxBitmap32.CreateSize(R);
      Painter.DrawScrollBarPart(Bmp.CxCanvas, Sender.Kind = sbHorizontal, R, sbpLineDown,
        SRVDrawStateToCxButtonState(State));
      Canvas.Draw(ARect.Left, ARect.Top, Bmp);
      DoDefault := False;
    end;
  except
  end;
  Bmp.Free;
end;

procedure TSRVDxSkinAdapter.DrawScrollBarThumbTab(Sender: TSRVScrollBar;
  Canvas: TCanvas; ARect, PaintRect: TRect; State: TSRVDrawStates; var DoDefault: Boolean);
var
  Painter: TcxCustomLookAndFeelPainter;
  R: TRect;
  Bmp: TcxBitmap32;
begin
  Bmp := nil;
  try
    Painter := GetPainter();
    if (Painter <> nil) then
    begin
      SubtractRect(R, Sender.BoundsRect, TSRVSBarAccess(Sender).GetRectButtonRight());
      IntersectRect(R, ARect, R);
      OffsetRect(R, -R.Left, -R.Top);
      Bmp := TcxBitmap32.CreateSize(R);
      Painter.DrawScrollBarPart(Bmp.CxCanvas, Sender.Kind = sbHorizontal, R, sbpThumbnail,
        SRVDrawStateToCxButtonState(State));
      Canvas.Draw(ARect.Left, ARect.Top, Bmp);
      DoDefault := False;
    end;
  except
  end;
  Bmp.Free;
end;

function TSRVDxSkinAdapter.GetCxCanvas(Canvas: TCanvas): TcxCanvas;
begin
  if (FCxCanvasCache = nil) or (FCxCanvasCache.Canvas <> Canvas) then
  begin
    FCxCanvasCache.Free;
    FCxCanvasCache := TcxCanvas.Create(Canvas);
  end;
  Result := FCxCanvasCache;
end;

function TSRVDxSkinAdapter.GetPainter: TcxCustomLookAndFeelPainter;
var
  CrtLookAndFeel: TcxLookAndFeel;
begin
  CrtLookAndFeel := RootLookAndFeel();
  if CrtLookAndFeel <> nil then
    Result := CrtLookAndFeel.Painter
  else
    Result := nil;
end;

procedure TSRVDxSkinAdapter.LookAndFeelChanged(Sender: TcxLookAndFeel;
  AChangedValues: TcxLookAndFeelValues);
begin
  AdjustControls;
  if Assigned(FSavLookAndFeelChanged) then
    FSavLookAndFeelChanged(Sender, AChangedValues);
end;

procedure TSRVDxSkinAdapter.Refresh;
begin
  SetPaintHandlers(True);
  AdjustControls;
end;

procedure TSRVDxSkinAdapter.ReleaseCxCanvas(Canvas: TcxCanvas);
begin
  { do nothing }
end;

procedure TSRVDxSkinAdapter.SetDrawEditorFrame(const Value: Boolean);
begin
  FDrawEditorFrame := Value;
end;

procedure TSRVDxSkinAdapter.SetLookAndFeelChangedHandler(Attach: Boolean);
var
  CrtLookAndFeel: TcxLookAndFeel;
  IsAttached: Boolean;
begin
  CrtLookAndFeel := RootLookAndFeel();
  if CrtLookAndFeel <> nil then
  begin
    IsAttached :=
      (TMethod(CrtLookAndFeel.OnChanged).Data = Self) and
      (TMethod(CrtLookAndFeel.OnChanged).Code = @TSRVDxSkinAdapter.LookAndFeelChanged);
    if Attach and not IsAttached then
    begin
      FSavLookAndFeelChanged := CrtLookAndFeel.OnChanged;
      CrtLookAndFeel.OnChanged := LookAndFeelChanged;
    end
    else if not Attach and IsAttached then
    begin
      CrtLookAndFeel.OnChanged := FSavLookAndFeelChanged;
      FSavLookAndFeelChanged := nil;
    end;
  end;
end;

procedure TSRVDxSkinAdapter.SetPaintHandlers(Attach: Boolean);
var
  I: Integer;
  Component: TComponent;
begin
  if Owner <> nil then
  begin
    for I := 0 to Owner.ComponentCount - 1 do
    begin
      Component := Owner.Components[I];
      if Component is TSRichViewEdit then
        SetSRichViewPaintHandlers(TSRichViewEdit(Component), Attach);
    end;
  end;
end;

procedure TSRVDxSkinAdapter.SetSRichViewPaintHandlers(SRichView: TSRichViewEdit;
  Attach: Boolean);
var
  IsAttached: Boolean;
begin
  IsAttached :=
    (TMethod(SRichView.OnPaint).Data = Self) and
    (TMethod(SRichView.OnPaint).Code = @TSRVDxSkinAdapter.SRichViewPaint);
  if Attach <> IsAttached then
  begin
    if Attach then
      SRichView.OnPaint := SRichViewPaint
    else
      SRichView.OnPaint := nil;
    if Assigned(SRichView.ScrollBarV) then
    begin
      if Attach then
      begin
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawArea := DrawScrollBarArea;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawThumbTab := DrawScrollBarThumbTab;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawIncButton := DrawScrollBarIncButton;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawDecButton := DrawScrollBarDecButton;
        TSRVSBarAccess(SRichView.ScrollBarV).ParentBackground := False;
      end
      else
      begin
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawArea := nil;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawThumbTab := nil;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawIncButton := nil;
        TSRVSBarAccess(SRichView.ScrollBarV).OnDrawDecButton := nil;
      end;
    end;
    if Assigned(SRichView.ScrollBarH) then
    begin
      if Attach then
      begin
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawArea := DrawScrollBarArea;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawThumbTab := DrawScrollBarThumbTab;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawIncButton := DrawScrollBarIncButton;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawDecButton := DrawScrollBarDecButton;
        TSRVSBarAccess(SRichView.ScrollBarH).ParentBackground := False;
      end
      else
      begin
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawArea := nil;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawThumbTab := nil;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawIncButton := nil;
        TSRVSBarAccess(SRichView.ScrollBarH).OnDrawDecButton := nil;
      end;
    end;
  end;
end;

procedure TSRVDxSkinAdapter.SRichViewPaint(Sender: TSRichViewEdit; Canvas: TCanvas;
  Prepaint: Boolean; PaintRect: TRect);
var
  Painter: TcxCustomLookAndFeelPainter;
  CxCanvas: TcxCanvas;
  I: Integer;
  ClientRect, PageRect, Borders, R: TRect;
  WindowOrg: TPoint;
  hSavClipRgn: HRGN;
  hRegion: HRGN;
begin
  if Prepaint and Sender.PageProperty.PageViewMode then
  begin
    Painter := GetPainter();
    if (Painter <> nil) then
    begin
      CxCanvas := GetCxCanvas(Canvas);
      if (CxCanvas <> nil) then
      begin
        ClientRect := Sender.ClientRect;
        hSavClipRgn := CreateRectRgn(0, 0, 1, 1);
        if GetClipRgn(Canvas.Handle, hSavClipRgn) <= 0 then begin
          DeleteObject(hSavClipRgn);
          hSavClipRgn := 0;
        end;
        GetWindowOrgEx(Canvas.Handle, WindowOrg);
        R := ClientRect;
        OffsetRect(R, -WindowOrg.X, -WindowOrg.Y);
        hRegion := CreateRectRgnIndirect(R);
        SelectClipRgn(Canvas.Handle, hRegion);
        DeleteObject(hRegion);
        Borders := Painter.PrintPreviewPageBordersWidth;
        for I := Sender.GetFirstVisiblePage to Sender.GetLastVisiblePage do
        begin
          PageRect := Sender.GetPageClientRect(I);
          OffsetRect(PageRect, -WindowOrg.X, -WindowOrg.Y);
          InflateRect(PageRect, -1, -1);
          with PageRect do
            ExcludeClipRect(Canvas.Handle, Left, Top, Right, Bottom);
        end;
        Painter.DrawPrintPreviewBackground(CxCanvas, ClientRect);
        for I := Sender.GetFirstVisiblePage to Sender.GetLastVisiblePage do
        begin
          PageRect := Sender.GetPageClientRect(I);
          if IntersectRect(R, PageRect, ClientRect) then
          begin
            R := PageRect;
            Dec(R.Left, Borders.Left);
            Dec(R.Top, Borders.Top);
            Inc(R.Right, Borders.Right);
            Inc(R.Bottom, Borders.Bottom);
            Painter.DrawPrintPreviewPageBackground(CxCanvas, R, PageRect,
              not Sender.ReadOnly and (I = Sender.CurrentPage), True);
          end;
        end;
        SelectClipRgn(Canvas.Handle, hSavClipRgn);
        if hSavClipRgn <> 0 then
          DeleteObject(hSavClipRgn);
        if FDrawEditorFrame then
          Painter.DrawBorder(CxCanvas, ClientRect);
        ReleaseCxCanvas(CxCanvas);
      end;
    end;
  end;
end;

function TSRVDxSkinAdapter.SRVDrawStateToCxButtonState(State: TSRVDrawStates): TcxButtonState;
begin
  if not (srvdsEnabled in State) then
    Result := cxbsDisabled
  else if srvdsDown in State then
    Result := cxbsPressed
  else if srvdsHot in State then
    Result := cxbsHot
  else
    Result := cxbsDefault;
end;

end.